from datetime import datetime
from io import BytesIO
from typing import Optional

import pandas as pd
from django.db import models
from pandas.tseries.offsets import YearEnd
from wbcore.contrib.io.backends import AbstractDataBackend, register

from .mixin import DataBackendMixin
from .utils import Controller

DEFAULT_MAPPING = {
    "IBEFPD": "period_type",  # Expected
    "IBQ1EEDT": "period_end_date",  # Expected Q1
    "IBQ1ERDT": "expected_report_date",  # Expected Q1
    "IBQ2EEDT": "period_end_date",  # Expected Q2
    "IBQ2ERDT": "expected_report_date",  # Expected Q2
    "IBFPD": "period_type",  # Company/Actual
    "IBQ1ENDT": "period_end_date",  # Company/Actual Q1
    "IBQ1CRDT": "expected_report_date",  # Company/Actual Q1
    "IBQ2ENDT": "period_end_date",  # Company/Actual Q2
    "IBQ2CRDT": "expected_report_date",  # Company/Actual Q1
}


@register("Fiscal Period", provider_key="refinitiv", save_data_in_import_source=False, passive_only=False)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    CHUNK_SIZE = 20

    def __init__(self, import_credential: Optional[models.Model] = None, **kwargs):
        self.controller = Controller(import_credential.username, import_credential.password)

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = execution_time.date()
        if obj_external_ids:
            df = self.controller.get_data(obj_external_ids, list(DEFAULT_MAPPING.keys()))
            if not df.empty:
                df_last_year_end = self.controller.get_data(
                    obj_external_ids,
                    ["WC05350", "WC05351"],
                    start=(execution_date - YearEnd(1)).date(),
                    end=(execution_date + YearEnd(0)).date(),
                    freq="Y",
                )
                # We do this to gather the last fiscal year end date to forward to the parser in order for figuring out what is the actual period index
                if not df_last_year_end.empty:
                    df_last_year_end = df_last_year_end.sort_values(by="WC05350").groupby(["Instrument"]).last()
                    df = pd.concat(
                        [df.set_index("Instrument"), df_last_year_end[df_last_year_end.columns.difference(["Dates"])]],
                        axis=1,
                    ).reset_index(names="Instrument")
                df = df.dropna(how="all", subset=df.columns.difference(["Instrument", "Dates"]))
                if not df.empty:
                    content_file = BytesIO()
                    df.to_json(content_file, orient="records")
                    file_name = f"fiscal_periods_{datetime.timestamp(execution_time)}.json"
                    yield file_name, content_file



from typing import Any, Dict, Optional

from django.db import models
from django_fsm import FSMField, transition
from wbcore.contrib.authentication.models import User
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.geography.models import Geography
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.tags.models import Tag
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton, ButtonDefaultColor
from wbcore.models import WBModel

from wbfdm.models.instruments.classifications import Classification

from .instruments import Instrument, InstrumentType


class InstrumentRequest(WBModel):
    class Status(models.TextChoices):
        PENDING = "PENDING", "Pending"
        APPROVED = "APPROVED", "Approved"
        DENIED = "DENIED", "Denied"
        DRAFT = "DRAFT", "Draft"

    status = FSMField(
        default=Status.DRAFT,
        choices=Status.choices,
        verbose_name="Status",
        help_text="The Request Status (default to Pending)",
    )
    requester = models.ForeignKey(
        "directory.Person",
        related_name="instrument_requests",
        on_delete=models.SET_NULL,
        verbose_name="Requester",
        null=True,
        blank=True,
    )
    handler = models.ForeignKey(
        "directory.Person",
        related_name="handled_instrument_requests",
        on_delete=models.SET_NULL,
        verbose_name="Handler",
        null=True,
        blank=True,
    )
    notes = models.TextField(null=True, blank=True, verbose_name="Notes")
    created = models.DateTimeField(auto_now_add=True, verbose_name="Created", help_text="The request creation time")
    instrument_data = models.JSONField(default=dict, verbose_name="Instrument Data")
    created_instrument = models.OneToOneField(
        "wbfdm.Instrument", on_delete=models.CASCADE, blank=True, null=True, related_name="creation_request"
    )

    def __str__(self):
        return f'Instrument Request - {self.Status[self.status].label} ({"".join([f"{k}={v}" for k, v in self.instrument_data.items()])})'

    @property
    def deserialize_instrument_data(self) -> tuple[Dict[str, Any], Dict[str, Any]]:
        instrument_data = self.instrument_data.copy()
        many_to_many_data = dict()
        if currency_id := instrument_data.get("currency", None):
            instrument_data["currency"] = Currency.objects.filter(id=currency_id).first()
        if country_id := instrument_data.get("country", None):
            instrument_data["country"] = Geography.countries.filter(id=country_id).first()
        if instrument_type_id := instrument_data.get("instrument_type", None):
            instrument_data["instrument_type"] = InstrumentType.objects.get(id=instrument_type_id)
        if tags_list := instrument_data.pop("tags", None):
            many_to_many_data["tags"] = [Tag.objects.filter(id=tag_id).first() for tag_id in tags_list]
        if classifications_list := instrument_data.pop("classifications", None):
            many_to_many_data["classifications"] = [
                Classification.objects.filter(id=classification_id).first()
                for classification_id in classifications_list
            ]
        instrument_data["is_investable_universe"] = True
        return instrument_data, many_to_many_data

    def _check_already_existing_instrument(self):
        return (
            (isin := self.instrument_data.get("isin", None)) and Instrument.objects.filter(isin=isin).exists()
        ) or (
            (ric := self.instrument_data.get("refinitiv_identifier_code", None))
            and Instrument.objects.filter(refinitiv_identifier_code=ric).exists()
        )

    @transition(
        field=status,
        source=[Status.PENDING],
        target=Status.APPROVED,
        permission=lambda instance, user: user.has_perm("wbfdm.administrate_instrument"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbfdm:instrumentrequest",),
                icon=WBIcon.APPROVE.icon,
                key="approve",
                label="Approve",
                action_label="Approve",
                description_fields="<p>You are sure to approve this request?</p>",
            )
        },
    )
    def approve(self, by: Optional[User] = None, **kwargs):
        deserialize_instrument_data, many_to_many_data = self.deserialize_instrument_data
        created_instrument = Instrument.objects.create(**deserialize_instrument_data)
        for key, items in many_to_many_data.items():
            getattr(created_instrument, key).set(items)
        if profile := getattr(by, "profile", None):
            self.handler = profile
        self.created_instrument = created_instrument

    def can_approve(self):
        if self._check_already_existing_instrument():
            return {"non_field_errors": "An instrument already exists with the proposed identifier"}

    @transition(
        field=status,
        source=[Status.PENDING],
        target=Status.DENIED,
        permission=lambda instance, user: user.has_perm("wbfdm.administrate_instrument"),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbfdm:instrumentrequest",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Deny",
                description_fields="<p>You are sure to deny this request?</p>",
            )
        },
    )
    def deny(self, by: Optional[User] = None, **kwargs):
        if profile := getattr(by, "profile", None):
            self.handler = profile

    @transition(
        field=status,
        source=[Status.DENIED],
        target=Status.DRAFT,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbfdm:instrumentrequest",),
                icon=WBIcon.SEND.icon,
                key="backtodraft",
                label="Back To Draft",
                action_label="Back To Draft",
                description_fields="<p>You are sure to put this request back to draft?</p>",
            )
        },
    )
    def backtodraft(self, by: Optional[User] = None, **kwargs):
        pass

    @transition(
        field=status,
        source=[Status.DRAFT],
        target=Status.PENDING,
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                color=ButtonDefaultColor.WARNING,
                identifiers=("wbfdm:instrumentrequest",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label="Submit",
                action_label="Submit",
                description_fields="<p>You are sure to submit this request back?</p>",
            )
        },
    )
    def submit(self, by: Optional[User] = None, **kwargs):
        pass

    def can_submit(self) -> Dict[str, Any]:
        if self._check_already_existing_instrument():
            return {"non_field_errors": "An instrument already exists with the proposed identifier"}

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbfdm:instrumentrequestrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "Request {{id}} ({{status}})"

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbfdm:instrumentrequest"



from datetime import date

from django.db import models
from pandas.tseries.offsets import BYearEnd
from psycopg.types.range import DateRange
from wbcore import filters as wb_filters

from wbfdm.figures.financials.financial_analysis_charts import (
    PeriodChoices,
    VariableChoices,
)
from wbfdm.models.instruments.instruments import Instrument


def byearend_n_year_ago(n):
    today = date.today()
    return (today - BYearEnd(n)).date()


def byearend_2_year_ago(field, request, view):
    return byearend_n_year_ago(2)


class GroupKeyFinancialsFilterSet(wb_filters.FilterSet):
    group_keys = wb_filters.CharFilter(
        required=True,
        method=lambda q, n, v: q,
    )

    class Meta:
        models = Instrument
        fields = {}


class FinancialAnalysisFilterSet(wb_filters.FilterSet):
    class Meta:
        model = Instrument
        fields = {}


def _get_12m(field, request, view):
    return date(date.today().year - 1, 1, 1)


class FinancialAnalysisValuationRatiosFilterSet(wb_filters.FilterSet):
    date = wb_filters.FinancialPerformanceDateRangeFilter(
        method=lambda queryset, label, value: queryset,
        label="Date Range",
        required=True,
        clearable=False,
        default=lambda r, v, q: DateRange(_get_12m(r, v, q), date.today()),
    )

    class OutputChoices(models.TextChoices):
        TSTABLE = "TSTABLE", "Table (Time-series)"
        TABLE = "TABLE", "Table (Last Value)"
        CHART = "CHART", "Chart"

    class RangeChoices(models.TextChoices):
        MINMAX = "MINMAX", "Min-Max (entire period)"
        ROLLING = "ROLLING", "Rolling"

    output = wb_filters.ChoiceFilter(
        choices=OutputChoices.choices, label="Output", method="fake_filter", default=OutputChoices.CHART
    )
    period = wb_filters.ChoiceFilter(
        choices=PeriodChoices.choices, label="Period", method="fake_filter", default=PeriodChoices.NTM
    )
    vs_related = wb_filters.BooleanFilter(label="Versus related", default=False, method="fake_filter")
    clean_data = wb_filters.BooleanFilter(label="Clean data", default=True, method="fake_filter")
    ranges = wb_filters.BooleanFilter(label="Draw ranges", default=False, required=False, method="fake_filter")
    range_type = wb_filters.ChoiceFilter(
        choices=RangeChoices.choices,
        label="Range type",
        method="fake_filter",
        required=True,
        default=RangeChoices.MINMAX,
    )
    range_period = wb_filters.NumberFilter(
        precision=0, label="Rolling period", method="fake_filter", required=True, default=120
    )
    x_axis_var = wb_filters.ChoiceFilter(
        choices=VariableChoices.choices, label="X-Axis", method="fake_filter", default=VariableChoices.EPSG
    )
    y_axis_var = wb_filters.ChoiceFilter(
        choices=VariableChoices.choices, label="Y-Axis", method="fake_filter", default=VariableChoices.PE
    )
    z_axis_var = wb_filters.ChoiceFilter(
        choices=VariableChoices.choices, label="Bubble", method="fake_filter", default=VariableChoices.MKTCAP
    )
    median = wb_filters.BooleanFilter(label="Median", default=True, method="fake_filter")

    class Meta:
        model = Instrument
        fields = {}


class EarningsAnalysisFilterSet(wb_filters.FilterSet):
    date = wb_filters.FinancialPerformanceDateRangeFilter(
        method=lambda queryset, label, value: queryset,
        label="Date Range",
        required=True,
        clearable=False,
        default=lambda r, v, q: DateRange(_get_12m(r, v, q), date.today()),
    )

    class OutputChoices(models.TextChoices):
        EPS = "EPS", "Earnings ($)"

    analysis = wb_filters.ChoiceFilter(
        choices=OutputChoices.choices, label="Analysis", method=lambda q, n, v: q, default=OutputChoices.EPS
    )
    period = wb_filters.ChoiceFilter(
        choices=PeriodChoices.choices, label="Period", method=lambda q, n, v: q, default=PeriodChoices.NTM
    )
    vs_related = wb_filters.BooleanFilter(label="Show related", default=False, method=lambda q, n, v: q)

    class Meta:
        model = Instrument
        fields = {}



from unittest.mock import patch

import pytest
from django.contrib.auth.models import Permission
from faker import Faker
from wbcore.contrib.io.exceptions import DeserializationError

from wbfdm.import_export.handlers.instrument_list import InstrumentListImportHandler
from wbfdm.models import InstrumentListThroughModel

fake = Faker()


@pytest.mark.django_db
class TestInstrumentListModel:
    def test_instrument_list_creation(self, instrument_list):
        assert instrument_list.id is not None

    @pytest.mark.parametrize(
        "from_date,to_date,comment",
        [
            (fake.past_date(), fake.future_date(), fake.sentence()),
            (fake.past_date(), fake.future_date(), None),
            (fake.past_date(), None, fake.sentence()),
            (None, fake.future_date(), fake.sentence()),
        ],
    )
    def test_handler_deserialization(self, instrument_list, instrument, import_source, from_date, to_date, comment):
        data = dict()
        if from_date:
            data["from_date"] = from_date.strftime("%Y-%m-%d")
        if to_date:
            data["to_date"] = to_date.strftime("%Y-%m-%d")
        if comment:
            data["comment"] = comment

        data["instrument"] = instrument.id
        data["instrument_list"] = instrument_list.id
        handler = InstrumentListImportHandler(import_source)
        handler._deserialize(data)
        assert data["instrument"] == instrument
        assert data["instrument_list"] == instrument_list
        if to_date:
            assert data["to_date"] == to_date
        if from_date:
            assert data["from_date"] == from_date
        if comment:
            assert data["comment"] == comment
        # try with a dictionary
        data = {}
        data["instrument"] = {"instrument_type": instrument.instrument_type, "isin": instrument.isin}
        data["instrument_list"] = {
            "identifier": instrument_list.identifier,
            "name": instrument_list.name,
            "instrument_list_type": instrument_list.instrument_list_type,
        }
        handler._deserialize(data)
        assert data["instrument"] == instrument
        assert data["instrument_list"] == instrument_list

        # assert missing instrument raise error
        data = {}
        data["instrument"] = instrument.id
        with pytest.raises((DeserializationError,)):
            handler._deserialize(data)
            raise DeserializationError

        # automatically match the instrument if the string representation is provided and the match has already been done for that list
        data = {}
        data["instrument_list"] = instrument_list.id
        data["instrument_str"] = instrument.name
        InstrumentListThroughModel.objects.create(
            instrument_list=instrument_list, instrument_str=instrument.name, instrument=instrument
        )
        handler._deserialize(data)
        assert data["instrument"].id == instrument.id

    @patch("wbfdm.import_export.handlers.instrument_list.send_notification")
    def test_post_processing_objects(
        self, mock_fct, import_source, instrument_list_factory, instrument_factory, user_factory
    ):
        initial_call = mock_fct.call_count
        pm = user_factory.create()
        user_factory.create()  # normal user

        pm.user_permissions.add(
            Permission.objects.get(content_type__app_label="wbfdm", codename="administrate_instrumentlist")
        )
        handler = InstrumentListImportHandler(import_source)

        # Test if the leftover elements are removed on every imports. (e.g. instrument that are not in an exclusion list anymore)
        list1 = instrument_list_factory.create()
        list2 = instrument_list_factory.create()

        leftover1 = InstrumentListThroughModel.objects.create(
            instrument_list=list1, instrument=instrument_factory.create()
        )
        remaining_obj1 = InstrumentListThroughModel.objects.create(
            instrument_list=list1, instrument=instrument_factory.create()
        )
        leftover2 = InstrumentListThroughModel.objects.create(
            instrument_list=list2, instrument=instrument_factory.create()
        )
        remaining_obj2 = InstrumentListThroughModel.objects.create(
            instrument_list=list2, instrument=instrument_factory.create()
        )

        handler._post_processing_objects([remaining_obj1, remaining_obj2], [], [])
        with pytest.raises((InstrumentListThroughModel.DoesNotExist)):
            leftover1.refresh_from_db()
        with pytest.raises((InstrumentListThroughModel.DoesNotExist)):
            leftover2.refresh_from_db()
        remaining_obj2.refresh_from_db()
        remaining_obj1.refresh_from_db()
        assert remaining_obj1
        assert remaining_obj2

        assert mock_fct.call_count - initial_call == 2  # 2 because there are two instrument added to the list
